/*
 * Copyright (C) 2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.acidhouse.jazzmaster;

import org.eiichiro.acidhouse.DatastoreSession;
import org.eiichiro.jazzmaster.bus.Around;
import org.eiichiro.jazzmaster.bus.Interceptor;
import org.eiichiro.jazzmaster.service.Consumes;
import org.eiichiro.reverb.reflection.Invocation;

/**
 * {@code TransactionalInterceptor} is a Jazzmaster interceptor component to 
 * intercept transactional method ({@code @Transactional} annotated Jazzmaster 
 * service interface method) and to control transaction commitment of Acid House 
 * {@code DatastoreSession}.
 * 
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
@Transactional
@Interceptor
public class TransactionalInterceptor {

	@Consumes DatastoreSession session;
	
	@Consumes @Strong DatastoreSession strongSession;

	/**
	 * Controls Acid House transaction.
	 * First, this method begins transaction on {@code AppEngineDatastoreSession} 
	 * and {@code AppEngineStrongDatastoreSession}. If the transactional method 
	 * invocation is succeeded, this method commits these transactions. If the 
	 * transactional method throws any exception, this method rolls back these 
	 * transactions and re-throws the exception.
	 * 
	 * @param invocation The encoded method invocation of transactional method.
	 * @return The invocation result.
	 * @throws Throwable If transactional method throws any exception.
	 */
	@Around
	public Object transact(Invocation<?> invocation) throws Throwable {
		session.begin();
		strongSession.begin();
		
		try {
			Object result = invocation.proceed();
			session.commit();
			strongSession.commit();
			return result;
		} catch (Throwable t) {
			session.rollback();
			strongSession.rollback();
			throw t;
		}
	}
	
}
